import processing.core.*; 
import processing.xml.*; 

import ddf.minim.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class game45_wavespark_ii extends PApplet {


/**
 * Game 45: "Wavespark II" <br/>
 * Hold space for gravity, tap space to ping <br/>
 * M = mute, P = pause, B = toggle background, R = toggle reflections<br/><br/>
 *  Background is an edited version of <a href="http://www.eso.org/public/images/potw1015a/">Stellar Titans of Pismis 24</a> by ESO/IDA/Danish 1.5 m/ R. Gendler, U.G. J\u00f8rgensen, J. Skottfelt, K. Harps\u00f8e<br/>used under the Creative Commons <a href="http://creativecommons.org/licenses/by/3.0/deed.en">Attribution 3.0 Unported</a> license.<br/>
 *  <br/><a href="http://nmccoy.net/">Go play more games at NMcCoy.net!</a><br/>
 * <a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/3.0/us/"><img alt="Creative Commons License" style="border-width:0" src="http://i.creativecommons.org/l/by-nc-sa/3.0/us/80x15.png" /></a><br />This <span xmlns:dc="http://purl.org/dc/elements/1.1/" href="http://purl.org/dc/dcmitype/InteractiveResource" rel="dc:type">work</span> by <a xmlns:cc="http://creativecommons.org/ns#" href="http://nmccoy.net" property="cc:attributionName" rel="cc:attributionURL">Nathan McCoy</a> is licensed under a <a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/3.0/us/">Creative Commons Attribution-Noncommercial-Share Alike 3.0 United States License</a>.
 */
String game_name = "Wavespark II";
boolean[] keys = new boolean[256];
boolean title;
boolean paused;
boolean muted;
boolean game_over;

boolean reflections = true;
boolean draw_bg = true;
int score;
int top_score;

int drawing_ms_last;
int physics_ms_last;
int CORE_FPS = 540;
int DRAWING_FPS = 90;
int PHYSICS_FPS = 180;
int DRAW_MS = 1000 / DRAWING_FPS;
int PHYS_MS = 1000 / PHYSICS_FPS;
Minim minim;

float WATER_SPEED = 2;
float GLIDE_ANGLE = PI/10;
float MAX_SKIP_ANGLE = PI/4;///(PI/3 + PI/4)/2; //because I'm too tired to do math in my head right now
float WATER_ACCEL = 0.005f;
float GLIDE_DRAG = 0.01f;
float WATER_OMEGA = 0.015f;
float SKIP_SPEED_MULT = 0.90f;
float WATER_OVERSPEED_LIMIT = 7;
float WATER_OVERSPEED_DRAG = 0.1f;

int SKIP_WINDOW = 99945;
int SKIP_PAST_WINDOW = 1;
float SKIP_MIN_SPEED = 3;

float WATER_LEVEL = 240;
float SWIM_UP = -PI/2 + 0.1f;
float SWIM_DOWN = -PI/5; //PI/2 - 0.1;
float DIVE_GRAVITY = 0.04f;
float DRIFT_GRAVITY = 0.01f;

float RACE_START_XSPEED = 5.0f;

float speed;
float angle;
PVector player_pos;
PVector player_vel;
PVector cam;
boolean dive;
int dive_time;
int rise_time;
int water_time;
int poom_timer;
int POOM_TIME = 60;
int POOM_COOLDOWN = 60;
int phys_frame;
int BOOST_TIME = 120;
float BOOST_PER_STAR = 0.6f;
float MIN_STAR_BOOST = 5.0f;
int boost_timer;

float invisiwall;
float INVISIWALL_DISTANCE = 4000;

float RING_RADIUS = 250;
float RING_LEAD = 1500;
float RING_SPACING = 700;
float last_ring;
float RING_MIN_Y = - 700;
float RING_MAX_Y = WATER_LEVEL - RING_RADIUS*2;

float CHECKPOINT_OFFSET = RING_LEAD + RING_SPACING/2;
float CHECKPOINT_SPACING = 35000;
float checkpoint;
boolean show_distance;
boolean racing;
int race_time;
int best_race;

float RING_MIN_BOOST = 4;
float RING_BOOST = 0.3f;

float STAR_RADIUS = 40;

float FIREWORK_SPEED = 2.5f;

int star_chain;

int star_bonus;
int alt_bonus;
int focus_bonus;
int most_stars;

int burst_timer;
int BURST_TIME = PHYSICS_FPS * 4;
int race_echo_timer;



ParticleList particles = new ParticleList();
GameList rings = new GameList();

AudioSample diveSound;
AudioSample surfSound;
AudioSample skipSound;
AudioSample poomSound;
AudioSample aceSound;
AudioSample boingSound;
AudioSample burstSound;
AudioSample grabSound;
AudioSample checkSound;

public void audioInit()
{
  minim = new Minim(this);
  diveSound = minim.loadSample("dive.wav", 1024);
  surfSound = minim.loadSample("surface.wav", 1024);
  skipSound = minim.loadSample("skip.wav", 1024);
  poomSound = minim.loadSample("poom.wav", 1024);
  aceSound = minim.loadSample("ace.wav", 1024);
  boingSound = minim.loadSample("boing.wav", 1024);
  burstSound = minim.loadSample("starburst.wav", 1024);
  grabSound = minim.loadSample("grab.wav", 1024);
  checkSound = minim.loadSample("checkpoint.wav", 1024);
    
}
public void audioClose()
{
  diveSound.close();
  surfSound.close();
  skipSound.close();
    poomSound.close();
    aceSound.close();
    boingSound.close();
    burstSound.close();
    grabSound.close();
  checkSound.close();
  minim.stop();
  
}

PImage bg;

public void setup()
{
  size(720, 480);
  frameRate(600);
  audioInit();
  info_font = loadFont("PressStartK-16.vlw");
  setTitle();
  bg = loadImage("background6.jpg");
  bg.loadPixels();
  frame.setTitle(game_name);
}

public void setTitle()
{
  title = true;
  game_over = true;
}

public void newGame()
{
  game_over = false;
  title = false;
  paused = false;
  score = 0;
  player_pos = new PVector(0, 0);
  player_vel = new PVector(WATER_SPEED, 0);
  cam = new PVector(0, 0);
  dive_time = 0;
  rise_time = 0;
  water_time = 0;
  phys_frame = 0;
  star_chain = 0;
  burst_timer = 0;
  particles.clear();
  rings.clear();
  poom_timer = POOM_COOLDOWN;
  last_ring = RING_LEAD;
  invisiwall = -INVISIWALL_DISTANCE;
  checkpoint = CHECKPOINT_OFFSET + CHECKPOINT_SPACING;
  bgmLoop("wavespark2.mp3");
  boost_timer = 0;
  show_distance = false;
  racing = false;
  race_time = 0;
  best_race = MAX_INT;
}

public void stop()
{
  audioClose();
  super.stop();
}


public void populateRings()
{
  while(cam.x + RING_LEAD > last_ring)
  {
    rings.add(new Ring(new PVector(last_ring, random(RING_MAX_Y - RING_MIN_Y) + RING_MIN_Y)));
    last_ring += RING_SPACING;
  }
}

public void drawStarChain()
{
  if(star_chain > 1)
  {
    float theta = phys_frame * 0.02f;
    noStroke();
    colorMode(HSB);
    fill((phys_frame * 5)%256, 64, 255);
    for(int i = 0; i < star_chain; i++)
    {
      float angle = theta + i*TWO_PI/star_chain;
      float rad = (star_chain + 2) * 5;
      
      if(player_vel.mag() < FIREWORK_SPEED) rad = rad * player_vel.mag() / FIREWORK_SPEED;
      
      regStar(player_pos.x +  rad * cos(angle), player_pos.y + rad * sin(angle), 10, 5, angle + phys_frame * 0.02f);
    }
    colorMode(RGB);
  }
}

public void physicsStep()
{
  if(!title)
  {
    boolean old_dive = dive;
    dive = keys[' '];
    if(dive)
    {
      dive_time ++;
      rise_time = 0;
    } 
    else
    {
      rise_time ++;
      if(old_dive && dive_time < 30 && poom_timer >= POOM_COOLDOWN)
      {
        particles.add(new Poom());
        //if(player_vel.mag() < FIREWORK_SPEED)
        {
          boolean in_ring = false;
          for(Object o : rings)
          {
            GameObject g = (GameObject) o;
            if(inRadius(player_pos, g.pos, g.radius)) in_ring = true;
          }
          
          if(!in_ring && star_chain > 1) //do starburst 
          {
            if(!muted) burstSound.trigger();
            particles.add(new Starburst(star_chain));
            star_bonus = star_chain;
            alt_bonus = PApplet.parseInt(abs(player_pos.y - WATER_LEVEL) / 100) + 1;
            focus_bonus = max(PApplet.parseInt((1.0f + (1 - player_vel.mag()/FIREWORK_SPEED)*1.5f) * 100), 100);         
            score = star_bonus * alt_bonus * focus_bonus;
            burst_timer = BURST_TIME;
            
            boost_timer = BOOST_TIME;
            float speed = player_vel.mag();
            float angle = player_vel.heading2D();
            speed += BOOST_PER_STAR * (star_chain - 1);
            if(speed < MIN_STAR_BOOST) speed = MIN_STAR_BOOST;
            player_vel.x = cos(angle)*speed;
            player_vel.y = sin(angle)*speed;
                        
            star_chain = 0;
          }
          else
          {
            if(!muted) poomSound.trigger();
          }
          
        }
     //   else 
     //   {
     //     if(!muted) poomSound.trigger();
     //   }
        poom_timer = 0;
      }
      dive_time = 0;
    }

    //main game loop
    speed = player_vel.mag();
    angle = player_vel.heading2D();
    boolean underwater = (player_pos.y > WATER_LEVEL);
    if(underwater)
    {
      if(water_time < SKIP_PAST_WINDOW && !dive && abs(player_vel.y)<abs(player_vel.x) && speed > SKIP_MIN_SPEED)
      {
        //Water skip!
        if(!muted) skipSound.trigger();

        player_vel.y = -player_vel.y;
        player_pos.y = WATER_LEVEL;
        player_vel.mult(SKIP_SPEED_MULT);
        splash();
        //angle = player_vel.heading2D();
      }
      else
      {
        if(water_time == 0) 
        {
          if(!muted) diveSound.trigger();
          star_chain = 0;
          splash();
        }

        water_time++;
        if(dive) angle = angleTowards(angle, SWIM_DOWN, WATER_OMEGA);
        else angle = angleTowards(angle, SWIM_UP, WATER_OMEGA);

        speed = tend(speed, WATER_SPEED, WATER_ACCEL);
        if(speed > WATER_OVERSPEED_LIMIT) speed -= WATER_OVERSPEED_DRAG;
        player_vel = new PVector(cos(angle)*speed, sin(angle)*speed);
      }
    }
    else
    {
      if(water_time > 0)
      {
        if(!muted)surfSound.trigger();
        splash();
      }
      water_time = 0;
      if(dive) player_vel.y += DIVE_GRAVITY;
      else player_vel.y += DRIFT_GRAVITY;
    }

    if(phys_frame % 9 == 0) particles.add(new SimpleTrail());

    player_pos.add(player_vel);
    
    invisiwall = max(invisiwall, player_pos.x - INVISIWALL_DISTANCE);
    
    if(player_pos.x < invisiwall)
    {
      if(!muted) boingSound.trigger();
      particles.add(new Wall());
      player_pos.x = invisiwall;
      player_vel.x = abs(player_vel.x);
    }
    
    if(player_pos.x > checkpoint)
    {
      if(racing)
      {
        racing = false;
        best_race = min(best_race, race_time);
        for(int i=0; i<height; i+=5)
        {
          PVector pos = new PVector(checkpoint, cam.y+i);
          float distance = PVector.dist(pos, player_pos);
          particles.add(new CheckMote(pos, PVector.add(randomVector(1.6f),PVector.mult(player_vel, 3/(distance/100+1))), 360));
        }
        if(!muted) checkSound.trigger();
        race_echo_timer = BURST_TIME;
      }
      else
      {
        if(boost_timer > 0 || poom_timer < POOM_TIME || player_vel.x > RACE_START_XSPEED)
        {
          if(!muted) aceSound.trigger();
          racing = true;
          race_time = 0;
          
          for(int i=-height/2; i<height*1.5f; i+=5)
          {
            PVector pos = new PVector(checkpoint, cam.y+i);
            particles.add(new StartMote(pos, new PVector(random(player_vel.x)+player_vel.x, 0), 360));
          }
        }
        else particles.add(new Wall(checkpoint));
      }
      checkpoint += CHECKPOINT_SPACING;
      show_distance = true;
      
    }

    cam.x = player_pos.x - 100;
    float campad;
    campad = height/2;
    float altitude = max(0, -player_pos.y + WATER_LEVEL);
    campad /= (1 + (altitude / (height/2)));

    cam.y = min(0, player_pos.y - campad);
    
    
    particles.tick();
    
    populateRings();
    
    rings.tick();
    
    top_score = max(score, top_score);
    most_stars = max(star_chain, most_stars);
  
    //timers  
    if(racing) race_time++;
    poom_timer++;
    burst_timer--;
    if(boost_timer > 0) particles.add(new ShinyTrail());
    boost_timer--;
    phys_frame++;
    race_echo_timer--;
    if(phys_frame % 180 == 0) 
    {
      fps = fpscounter;
      fpscounter = 0;
    }
  }
  
  


  //jukebox stuff
  info_time --;
}

public void splash()
{
  for(int i=0; i < 40; i++)
  {
    particles.add(new Droplet());
  }
}

int frx = 0xDEADBEEF;


int fpscounter;
int fps;
public void drawingStep()
{
  fpscounter++;
  if(title)
  {
    drawTitle();
  }
  else
  { 
    //background(64,32,96);
    background(24,0,64);
    int bgw = bg.width;
    int bgh = bg.height;
    int cx =  PApplet.parseInt(cam.x * 0.2f);
    int cy =  PApplet.parseInt(cam.y *0.2f);
    while(cx < 0) cx += bgw;
    while(cy < 0) cy += bgh;
    int bgx = -(cx % bgw);
    int bgy = -(cy % bgh);

    if(draw_bg)
    {
      image(bg, bgx, bgy);
      image(bg, bgx+bgw, bgy);
      image(bg, bgx, bgy+bgh);
      image(bg, bgx+bgw, bgy+bgh);
    }
//    loadPixels();
//    for(int x = 0; x < width; x++)
//    {      
//      int u = (cx + x)%bgw;
//      for(int y = 0; y < height; y++)
//      {
//
//        int v = (cy + y)%bgh;
//        pixels[x + y*width] = bg.pixels[u + v*bgw];
//      }
//    }
//    updatePixels();

    fill(255);
    stroke(255);
    ellipseMode(RADIUS);
    pushMatrix();
    translate(-cam.x, -(int)cam.y);

    particles.draw();

    stroke(255, 96);
    if(racing)
    {
          colorMode(HSB);
          stroke((phys_frame * 5)%256, 64, 255);
    }
    strokeWeight(2);
    line(checkpoint, cam.y, checkpoint, cam.y+height);
    colorMode(RGB);
    
    fill(255);
    stroke(255, 255, (phys_frame/3%16)*16);
    if(poom_timer <= POOM_TIME) stroke(255);
    else if(poom_timer <= POOM_COOLDOWN)
    {
      noStroke();
    }
    strokeWeight(2);
    float r = 6;
    ellipse(player_pos.x, player_pos.y, r, r);
    
    drawStarChain();
    
    fill(0, 48, 92, 128);
    rectMode(CORNERS);
    strokeWeight(2);
    stroke(255, 230, 192, 96);
    noStroke();
    rect(cam.x - 1, WATER_LEVEL+1, cam.x+width + 1, WATER_LEVEL + height/2+1);

    rings.draw();
    //draw reflection


    popMatrix();
    //if(player_pos.y < WATER_LEVEL)
    if(reflections)
    {
      
      int waterline = (int)(WATER_LEVEL - cam.y);
      if(waterline < height)
      {
      loadPixels();

      strokeWeight(1);
      for(int y = waterline; y < height; y++)
      {
        float dispmag = (y - waterline)*0.15f + 1;
        float yscale = (y - height) / (height - waterline);
        yscale *= yscale;
        int xdisp = PApplet.parseInt(random(dispmag * 2 + 1)) - PApplet.parseInt(dispmag);//int(noise(yscale * y * 0.005, cam.x * 0.05) * dispmag);
        int chance = (y - waterline)*6;
        for(int x = 0; x < width; x++)
        {
          frx = ((frx << 1) + (((frx >> 27) ^ (frx >> 30)) & 1));
          
//        float rn = random(1);
//           rn = rn * rn;
//           
//           PVector p = new PVector(random(width), rn*(height/2) + waterline);
//           PVector mp = p.get();
//           mp.y = waterline*2 - mp.y;
//           if(p.y < height)
//           {
//           stroke(pixels[(int)mp.x + (int)mp.y * width]);
//           float w = 5 + 0.05*(p.y - waterline);
//           line(p.x - w, p.y, p.x+w, p.y);
//           }
           
          int refy = waterline * 2 - y;
          int refx = x + xdisp;
          if(refx < 0) refx = 0;
          if(refx > width-1) refx = width - 1;
          if((frx & 0xFF) > chance)
            pixels[x + y*width] = pixels[refx + refy*width];
          else if((frx & 0xFF) > chance>>2)
            pixels[x + y*width] = 0xFF000000 | ((pixels[refx + refy*width]&0xFEFEFEFE) / 2 + (pixels[x + y*width]&0xFEFEFEFE) / 2);
        }
      }
      updatePixels();
      }
    }

    strokeWeight(2);
    stroke(255, 230, 192, 96);
    pushMatrix();
    translate(-(int)cam.x, -(int)cam.y);
    line(cam.x, WATER_LEVEL, cam.x+width, WATER_LEVEL);
    popMatrix();

    fill(255, 128);
    textFont(info_font, 16);
    textAlign(LEFT, BOTTOM);
    //text("fps: "+(int)fps, 8, height-8);

    if(burst_timer > 0)
    {
      fill(255, 255, 255 * burst_timer / BURST_TIME, 255 * burst_timer / BURST_TIME);
      textAlign(CENTER, BOTTOM);
      int fake_score = min(score, score * 4 - score * 4 * burst_timer / BURST_TIME);
      text("STARS \u00d7 "+star_bonus+"\nALTITUDE \u00d7 "+alt_bonus+"\nFOCUS \u00d7 "+focus_bonus, width/2, height/2);
      textFont(info_font, 32);
      textAlign(CENTER, TOP);
      text(fake_score, width/2, height/2 + 16);
    }
    
    textAlign(RIGHT, TOP);
    textFont(info_font, 8);
    fill(255, 128);
    {
      if(most_stars > 1)
      text("BEST STARS: "+most_stars, width-8, 8);
      if(top_score > 0)
      text("BEST BURST: "+top_score, width-8, 16);
      if(best_race < MAX_INT)
      {
        int tenths = (best_race/18) % 10;
        int secs = best_race / PHYSICS_FPS;
        text("BEST RACE: "+secs+"."+tenths, width-8, 24);
      }
    }
    if(show_distance)
    {
//      textFont(info_font, 8);
//      textAlign(RIGHT, BOTTOM);
//      text(int(checkpoint - player_pos.x)/100 + ">", min(width-8 - (checkpoint - player_pos.x)/CHECKPOINT_SPACING * (width - 60), checkpoint - cam.x - 8), height-8);
//      
      strokeWeight(1);
      stroke(255, 64);
      noFill();
      line(8, height-8, width-8, height-8);
      ellipse(8 + (width-16)*(1 - (checkpoint - player_pos.x)/CHECKPOINT_SPACING), height-8, 3, 3);
    }
    if(racing)
    {
      textFont(info_font, 16);
      textAlign(LEFT, TOP);
      int tths = PApplet.parseInt(abs(player_vel.x) * 10) % 10;
      int ispd = PApplet.parseInt(abs(player_vel.x));    
      text(ispd+"."+tths, 8, 16);
      textFont(info_font, 8);
      text("SPEED", 8, 8);
    }
    
    if(racing || race_echo_timer > 0)
    {
      if(race_echo_timer > 0)
      {
        fill(255, 255, 255 * race_echo_timer / BURST_TIME,255 * race_echo_timer / BURST_TIME);
      }
      textFont(info_font, 32);
      textAlign(CENTER, TOP);
      int tenths = (race_time * 10 / PHYSICS_FPS)%10;
      int secs = race_time / PHYSICS_FPS;
      text(secs+"."+tenths, width/2, 8);
    }

    if(paused) drawPauseOverlay();
  }

  //jukebox stuff
  drawSongInfo();
}

public void drawTitle()
{
  background(128, 0, 128);
  image(bg, 0, 0);
  fill(255);
  textFont(info_font, 32);
  textAlign(CENTER, CENTER);
  text("Wavespark II", width/2, height/2);
  textFont(info_font, 16);
  text("press ENTER to begin", width/2, height*2/3);
  textFont(info_font, 8);
  text("SPACE = gravity / ping\nP = pause   M = mute   B = toggle bg   R = toggle reflections", width/2, height*3/4);
}

public void drawPauseOverlay()
{
  rectMode(CORNER);
  strokeWeight(1);
  noStroke();
  fill(0, 128);
  rect(0, 0, width, height);
  fill(255);
  textFont(info_font, 16);
  textAlign(CENTER, CENTER);
  text("[ PAUSED ]", width/2, height/2);
}



public void draw()
{
  while(physics_ms_last + PHYS_MS < millis())
  {
    if(!paused)
    {
      physicsStep();
    }
    physics_ms_last += PHYS_MS;
  }
  if(drawing_ms_last + DRAW_MS < millis())
  {
    drawingStep();
    drawing_ms_last = millis();
  }
}

public void keyPressed()
{
  if(keyCode < 256)
    if(!keys[keyCode]) 
    {
      keys[keyCode] = true;
      down(keyCode);
    }
}
public void keyReleased()
{
  if(keyCode < 256)
    if(keys[keyCode])
    {
      keys[keyCode] = false;
      up(keyCode);
    }
}
public void down(int theKey)
{
  println(theKey + " down");
  if(theKey == 'M') jbToggleMute();
  if(theKey == 'B') draw_bg = !draw_bg;
  if(theKey == 'R') reflections = !reflections;
  if(theKey == 'P' && !title) paused = !paused;
  if(theKey == ENTER && game_over) newGame();
}
public void up(int theKey)
{
  println(theKey + " up");
}

class CheckMote extends Particle
{
  CheckMote(PVector pos, PVector vel, int life)
  {
    super(pos, vel, life);
  }
  public void draw()
  {
    colorMode(HSB);
    noStroke();
    fill((phys_frame * 5)%256, 64, 255);
    ellipse(pos.x, pos.y, 3*life/maxlife + 0.5f, 3*life/maxlife + 0.5f);
    
    colorMode(RGB);
  }
}
class StartMote extends Particle
{
  StartMote(PVector pos, PVector vel, int life)
  {
    super(pos, vel, life);
  }
  public void draw()
  {
    
    noStroke();
    fill(255);
    ellipse(pos.x, pos.y, 3*life/maxlife + 0.5f, 3*life/maxlife + 0.5f);
  }
}

class Wall extends Particle
{
  Wall()
  {
    super(new PVector(invisiwall, 0), new PVector(0, 0), 30);
  }
  Wall(float x)
  {
    super(new PVector(x, 0), new PVector(0, 0), 30);
  }
  
  public void draw()
  {
    stroke(255, 255*life/maxlife);
    strokeWeight(2);
    
    line(pos.x, cam.y, pos.x, cam.y+height);
  }
}

class ShinyTrail extends Particle
{
  ShinyTrail()
  {
    super(PVector.add(player_pos, randomVector(5)), new PVector(0, 0), 60);
  }
  public void draw()
  {
    colorMode(HSB);
    noStroke();
    fill((phys_frame * 5)%256, 64, 255);
    ellipse(pos.x, pos.y, 10 * life / maxlife, 10 * life / maxlife);    
    colorMode(RGB);
  }
}
class SimpleTrail extends Particle
{
  SimpleTrail()
  {
    super(PVector.add(player_pos, randomVector(3)), new PVector(0, 0), 60);
  }
  public void draw()
  {
    noFill();
    stroke(255);
    strokeWeight(1);
    if(pos.y < WATER_LEVEL)
    {
      stroke(255, 255, random(255));
      float r = 5.0f*life/maxlife;
      line(pos.x-r, pos.y, pos.x+r, pos.y);
      line(pos.x, pos.y-r, pos.x, pos.y+r);
    }
    else
    {
      stroke(random(255), 255, 255);
      ellipse(pos.x, pos.y, 5 * life / maxlife, 5 * life / maxlife);
    }
  }
}
class Poom extends Particle
{
  Poom()
  {
    super(player_pos, new PVector(0, 0), 30);
  }
  public void draw()
  {
    noFill();
    noStroke();
    fill(255, 255*life/maxlife);
    //strokeWeight(2);
    float r = 40 - 40*life/maxlife;
    ellipse(player_pos.x, player_pos.y, r, r);
  }
}

class Droplet extends Particle
{
  Droplet()
  {
    super(PVector.add(player_pos, randomVector(3)), PVector.add(PVector.mult(player_vel, random(2)), randomVector(random(player_vel.mag()/8))), 180); 
    vel.mult(1);
  }
  public void tick()
  {
    if(pos.y > WATER_LEVEL)
    {
      vel.y -= 0.3f;
      vel.mult(0.95f);
    }
    else
    {
      vel.y += 0.015f;
      vel.mult(0.99f);
    }
    super.tick();
  }
  public void draw()
  {
    noStroke();
    fill(255, 255*(2*life/maxlife + random(2) - 1));
    ellipse(pos.x, pos.y, 2, 2);
  }
}

class StarGrab extends Particle
{
  StarGrab(PVector pos)
  {
    super(pos, new PVector(0, 0), 60);
  }
  public void draw()
  {
    noStroke();
    colorMode(HSB);
    fill((phys_frame * 5)%256, 64, 255);
    ellipse(pos.x, pos.y, STAR_RADIUS * life / maxlife + 5, STAR_RADIUS * life / maxlife + 5);    
    colorMode(RGB);
  }
  public void tick()
  {
    PVector to_player = PVector.sub(player_pos, pos);
    to_player.div(life/4 + 1);
    pos.add(to_player);
    super.tick();
  }
}

class Starburst extends Particle
{
  int spokes;
  Starburst(int sides)
  {
    super(player_pos, new PVector(0, 0), 180);
    spokes = sides;
  }
  public void draw()
  {
    colorMode(HSB);
    stroke((phys_frame * 5)%256, 64, 255);
    strokeWeight(10 * life / maxlife);
    float theta = phys_frame * 0.02f;
    float radi = 3000 - 3000*life/maxlife;
    for(int i = 0; i < spokes; i++)
    {
      float angle = theta + i*TWO_PI/spokes;
      line(pos.x, pos.y, pos.x + radi * cos(angle), pos.y + radi * sin(angle));
    }
    noFill();
    ellipse(pos.x, pos.y, radi/2, radi/2);
    colorMode(RGB);
  }
}

class Ring extends GameObject
{
  boolean tagged;
  Ring(PVector p)
  {
    pos = p.get();
    radius = RING_RADIUS;
    tagged = false;
  }
  public void tick()
  {
    if(inRadius(player_pos, pos, radius)) 
    {
      tagged = true;
      if(poom_timer == 0)
      {
        //removal_flag = true;
        //if(!muted) aceSound.trigger();
        
        float speed = player_vel.mag();
        
        speed += RING_BOOST;
        if(speed < RING_MIN_BOOST) speed = RING_MIN_BOOST;
        
        PVector toward = PVector.sub(pos, player_pos);
        toward.normalize();
        toward.mult(speed);
        player_vel = toward;
        
//        player_vel.y = min(-abs(player_vel.y) - 0.5, -1.5);
//       player_vel.x += 0.5;
      }
    }
    else
    {
      //if(tagged) removal_flag = true;
    }
    if(inRadius(player_pos, pos, STAR_RADIUS) && poom_timer <= POOM_TIME)
    {
      removal_flag = true;
      particles.add(new StarGrab(pos));
      if(!muted) grabSound.trigger();
      star_chain++;
    }
    
    //if(pos.x + radius < cam.x - RING_LEAD) removal_flag = true;
    if(pos.x < invisiwall) removal_flag = true;
  }
  public void draw()
  {
    noFill();
    strokeWeight(2);
    colorMode(HSB);
    stroke((phys_frame * 5)%256, 64, 255, 128);
    ellipse(pos.x, pos.y, radius, radius);
    noStroke();
    fill((phys_frame * 5)%256, 64, 255);
    regStar(pos.x, pos.y, STAR_RADIUS, 5, phys_frame * 0.02f);
    colorMode(RGB);
  }
}
public void regStar(float x, float y, float r, int sides, float rot)
{
  beginShape();
  for(int i = 0; i < sides; i++)
  {
    float segAngle = TWO_PI/sides * 2;
    vertex(cos(i*segAngle + rot) * r + x,  sin(i*segAngle + rot) * r + y);
  }
  endShape(CLOSE);
}


AudioPlayer player;
String song_info;
int info_time;
int INFO_DURATION = PHYSICS_FPS * 6;
int INFO_TRANSITION = PHYSICS_FPS / 2;
PFont info_font;

public void jbToggleMute()
{
  muted = !muted;
  if(player != null)
  {
    if(muted) player.mute();
    else player.unmute();
  }
}

public void bgmDuck(float amount)
{
  if(player != null)
  {
    player.setGain(amount);
  }
}

public void bgmLoop(String filename)
{
  if(player != null) player.close();
  player = minim.loadFile(filename, 2048);
  if(muted) player.mute();
  if(player == null)
  {
    System.err.println("Not found: "+filename);
    return;
  }
  player.loop();
  AudioMetaData md = player.getMetaData(); 
  song_info = "\u266a "+ md.title() + " - "+md.author();
  info_time = INFO_DURATION;
}
public void bgmStop()
{
  if(player != null)
  {
     player.pause();
  }
}
public void bgmOnce(String filename)
{
   if(player != null) player.close();
  player = minim.loadFile(filename, 1024);
    if(muted) player.mute();
  if(player == null)
  {
    System.err.println("Not found: "+filename);
    return;
  }
  player.play();
  AudioMetaData md = player.getMetaData(); 
  song_info = "\u266a "+ md.title() + " - "+md.author();
  info_time = INFO_DURATION;
}

public void drawSongInfo()
{
  if(info_time <= 0) return;
  
  textFont(info_font, 8);
  int offset = 0;
  if(info_time < INFO_TRANSITION) offset = (int) map(info_time, 0, INFO_TRANSITION, 24, 0);
  if(info_time > INFO_DURATION - INFO_TRANSITION)
    offset = (int) map(info_time, INFO_DURATION - INFO_TRANSITION, INFO_DURATION, 0, 24);
  textAlign(LEFT, BOTTOM);
  fill(0, 128);
  /*
  text(song_info, 8, height-7+offset);
  text(song_info, 8, height-9+offset);
  text(song_info, 7, height-8+offset);
  text(song_info, 9, height-8+offset);*/
  rectMode(CORNER);
  noStroke();
  rect(0, height-18+offset, song_info.length()*8+16, 12);
  fill(192, 255, 192);
  text(song_info, 8, height-8+offset);
}
class GameList extends ArrayList
{
  public void tick()
  {
    Iterator i = iterator();
    while(i.hasNext())
    {
      GameObject p = (GameObject) i.next();
      p.tick();
      if(p.removal_flag) i.remove();
    }
  }
  public void draw()
  {
    Iterator i = iterator();
    while(i.hasNext())
    {
      GameObject p = (GameObject) i.next();
      p.draw();
    }
  }
}

abstract class GameObject
{
  public abstract void tick();
  public abstract void draw();
  PVector pos;
  float radius;
  boolean removal_flag = false;
}

class ParticleList extends ArrayList
{
  public void tick()
  {
    Iterator i = iterator();
    while(i.hasNext())
    {
      Particle p = (Particle) i.next();
      p.tick();
      if(p.life <= 0) i.remove();
    }
  }
  public void draw()
  {
    Iterator i = iterator();
    while(i.hasNext())
    {
      Particle p = (Particle) i.next();
      p.draw();
    }
  }
}

public boolean inRadius(PVector a, PVector b, float rad)
{
  return rad*rad > (a.x - b.x)*(a.x - b.x) + (a.y - b.y) * (a.y - b.y);// + (a.z - b.z) * (a.z - b.z);
}

abstract class Particle
{
  PVector pos;
  PVector vel;
  PVector accel;
  int life;
  int maxlife;
  Particle(PVector p, PVector v, int lifetime)
  {
    this(p, v, new PVector(0, 0), lifetime);
  }
  Particle(PVector p, PVector v, PVector a, int lifetime)
  {
    accel = new PVector(a.x, a.y);
    pos = new PVector(p.x, p.y);
    vel = new PVector(v.x, v.y);
    life = maxlife = lifetime;
  }
  public void tick()
  {
    vel.add(accel);
    pos.add(vel);
    life--;
  }
  public abstract void draw();
}

public PVector proj(PVector src, PVector dest)
{
  return PVector.mult(dest, src.dot(dest)/dest.dot(dest));
}

public void regPoly(float x, float y, float r, int sides, float rot)
{
  beginShape();
  for(int i = 0; i < sides; i++)
  {
    float segAngle = TWO_PI/sides;
    vertex(cos(i*segAngle + rot) * r + x,  sin(i*segAngle + rot) * r + y);
  }
  endShape(CLOSE);
}
public PVector randomVector(float magn)
{
  float angle = random(TWO_PI);
  return new PVector(cos(angle)*magn, sin(angle)*magn);
}

public float angleTowards(float current, float desired, float stepsize)
{
  float diff = desired - current;
  while(diff < 0) diff+=TWO_PI;
  while(diff >= TWO_PI) diff-=TWO_PI;
  if(diff < stepsize || diff > TWO_PI - stepsize) return desired;
  if(diff < PI) return current + stepsize;
  else return current - stepsize;
}

public int blendColors(int a, int b, float blent)
{
  return color(red(a)*(1-blent)+red(b)*(blent), green(a)*(1-blent)+green(b)*(blent), blue(a)*(1-blent)+blue(b)*(blent));
}

public void mosaic(int blocksize)
{
  if(blocksize < 1) blocksize = 1;
  loadPixels();
  for(int x = 0; x < width; x++)  for(int y = 0; y < height; y++)
  {
    pixels[x + y*width] = pixels[x/blocksize*blocksize + y/blocksize*blocksize*width];
  }
  updatePixels();
}
public void iris(int cx, int cy, int radius)
{
  loadPixels();
  for(int x = 0; x < width; x++)  for(int y = 0; y < height; y++)
  {
    if((x-cx)*(x-cx)+(y-cy)*(y-cy)>radius*radius)
      pixels[x + y*width] = color(0);
  }
  updatePixels();  
}
public void diamonds(float amount)
{  
  // background(255);
  loadPixels();
  for(int x = 0; x < width; x++) for(int y = 0; y < height; y++)
  {
    if((
    (abs(x-width/2)%32+abs(y-height/2)%32)+abs(x-width/2)/32*3+(abs(y-height/2)/32*3)
      ) > amount)
      pixels[x + y*width] = color(0);//color(0);
  }
  updatePixels();
}
public void drawCharge(float x, float y, float ratio, float radius, float circleSize, int numCircles)
{
  float angle = (ratio * TWO_PI);
  for(int i = 0; i<numCircles; i++)
  {
    ellipse(cos(angle)*radius*(1-ratio)+x, sin(angle)*radius*(1-ratio)+y, circleSize*(ratio), circleSize*(ratio));
    angle += TWO_PI/numCircles;
  }
}

public float tend(float input, float goal, float amount)
{
  if(abs(input - goal) < abs(amount)) return goal;
  if(input > goal) return input - abs(amount);
  return input + abs(amount);
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "game45_wavespark_ii" });
  }
}
